import datetime
import importlib
import json
import os
import re
import torch
import torch.distributed as dist
import torchaudio
from torchvision.io import write_video
from torchvision.transforms import ToTensor
from PIL import Image
from flask import Flask
from flask_cors import CORS
from flask_sock import Sock
import cv2
import numpy as np
import os
import requests
from tqdm import tqdm

def load_model_config(model_config_path=None):
    if model_config_path is None:
        model_config_path = os.path.join(os.getcwd(), "model.json")
    try:
        with open(model_config_path, "r") as file:
            model_config = json.load(file)
            default_params = model_config.get("params", {})
            plugin_name = model_config.get("name", "it broke")
    except FileNotFoundError:
        print(f"Warning: {model_config_path} not found. Using empty default parameters.")
        default_params = {}
        plugin_name = "it broke"
    return plugin_name, default_params


def check_for_saved_config(plugin_name, pipe_dict, default_params, load_pipe):
    model_dir = os.getenv("HF_HOME", "./models")
    filename = os.path.join(model_dir, f"{plugin_name}_server_config.json")
    os.makedirs(os.path.dirname(filename), exist_ok=True)  # Ensure the directory for the file exists
    if not os.path.exists(filename):
        with open(filename, 'w') as f:
            json.dump({}, f)
        data = {}
    else:
        with open(filename) as f:
            data = json.load(f)
    if data.get("default_params"):
        default_params.update(data["default_params"])
    if data.get("pipeline"): # this will be the name 
        pipe_dict["name"] = data["pipeline"]
        pipe_dict["pipeline"] = load_pipe(data["pipeline"])
    if data.get("loras"):
        pipe_dict["loras"] = data["loras"]
    if data.get("adapter_weights"):
        pipe_dict["adapter_weights"] = data["adapter_weights"]
    if "loras" in pipe_dict and len(pipe_dict["loras"]) > 0 and "adapter_weights" in pipe_dict and len(pipe_dict["adapter_weights"]) > 0:
        # convert the weights to numbers
        pipe_dict["adapter_weights"] = [float(weight) for weight in pipe_dict["adapter_weights"]]
        for lora in pipe_dict["loras"]:
            pipe_dict["pipeline"].load_lora_weights(lora['url'], weight_name=lora['weight_name'], adapter_name=lora['adapter_name'])
        pipe_dict["pipeline"].set_adapters([lora['adapter_name'] for lora in pipe_dict['loras']], adapter_weights=pipe_dict["adapter_weights"])
        pipe_dict["pipeline"].fuse_lora(adapter_names=[lora['adapter_name'] for lora in pipe_dict['loras']])


def save_different_params(plugin_name, params):
    model_dir = os.getenv("HF_HOME", "./models")
    filename = os.path.join(model_dir, f"{plugin_name}_server_config.json")
    try:
        with open(filename, "r") as f:
            current_config = json.load(f)
        saved_params = current_config.get("default_params", {})
        different_params = [k for k, v in saved_params.items() if k != 'prompt' and str(params.get(k)) != str(v)]
        if different_params:
            print(f"Different params: {', '.join(different_params)}")
            print("Saving new params to server config")
            # Update only the different parameters to avoid overwriting with defaults
            for param in different_params:
                current_config["default_params"][param] = str(params[param]) if isinstance(params[param], bool) else params[param]
            with open(filename, "w") as f:
                json.dump(current_config, f, indent=4)
    except FileNotFoundError:
        with open(filename, "w") as f:
            json.dump({"default_params": params}, f, indent=4)


# takes in a pillow file and saves it to disk with proper formatting
def save_image(file, file_ext, prefix):
    inference_directory = os.getenv("INFERENCE_DIRECTORY", "./inference")
    datestr = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    clean_prefix = re.sub(r'[^a-zA-Z0-9_-]', '', prefix)
    filename = f"{datestr}_{clean_prefix}.{file_ext}"
    outpath = os.path.join(inference_directory, filename)
    outdir = os.path.dirname(outpath)
    
    if not os.path.exists(outdir):
        os.makedirs(outdir)
    
    file.save(outpath, format='PNG')
    
    # Explicitly delete the image object if it's no longer needed
    del file
    return outpath

def custom_export_to_video(frames, out_path, fps=14):
    # Check the file extension
    file_ext = os.path.splitext(out_path)[1]
    if file_ext == '.gif':
        # Save frames as a GIF
        frames[0].save(out_path, save_all=True, append_images=frames[1:], loop=0, duration=1/fps)
    else:
        # Define the codec and create a VideoWriter object
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        video = cv2.VideoWriter(out_path, fourcc, fps, (frames[0].size[1], frames[0].size[0]))
        for frame in frames:
            # Convert the image from BGR to RGB
            frame = cv2.cvtColor(np.array(frame), cv2.COLOR_BGR2RGB)
            # Write the converted image
            video.write(frame)
        # Release the VideoWriter
        video.release()

def save_gif(frames, prefix, fps=24, quality=95, loop=1):
    imgs = frames #[Image.open(f) for f in sorted(frames)]
    if quality < 95:
        imgs = list(map(lambda x: x.resize((128, 128), Image.LANCZOS), imgs))
    inference_directory = os.getenv("INFERENCE_DIRECTORY", "./inference")
    datestr = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    # Clean the prefix by removing special characters other than "_" and "-"
    clean_prefix = re.sub(r'[^a-zA-Z0-9_-]', '', prefix)
    filename = f"{datestr}_{clean_prefix}.gif"
    outpath = os.path.join(inference_directory, filename) 
    outdir = os.path.dirname(outpath)
    duration_per_frame = 1000 // fps
    imgs[0].save(
        fp=outpath,
        format="GIF",
        append_images=imgs[1:],
        save_all=True,
        duration=duration_per_frame,
        loop=loop,
        quality=quality,
    )
    return outpath

def save_video(frames, prefix, fps=24, quality=95, audio_input=None):
    imgs = frames #[Image.open(f) for f in sorted(frames, key=lambda x: x.split("/")[-1])]
    if quality < 95:
        imgs = list(map(lambda x: x.resize((128, 128), Image.LANCZOS), imgs))

    img_tensors = [ToTensor()(img) for img in imgs]
    img_tensors = list(map(lambda x: x.unsqueeze(0), img_tensors))

    img_tensors = torch.cat(img_tensors)
    img_tensors = img_tensors * 255.0
    img_tensors = img_tensors.permute(0, 2, 3, 1)
    img_tensors = img_tensors.to(torch.uint8)

    inference_directory = os.getenv("INFERENCE_DIRECTORY", "./inference")
    datestr = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    # Clean the prefix by removing special characters other than "_" and "-"
    clean_prefix = re.sub(r'[^a-zA-Z0-9_-]', '', prefix)
    filename = f"{datestr}_{clean_prefix}.mp4"
    outpath = os.path.join(inference_directory, filename) 
    outdir = os.path.dirname(outpath)

    if audio_input is not None:
        audio_duration = len(img_tensors) / fps
        waveform, sr = torchaudio.load(audio_input)
        if waveform.shape[0] > 1:  # Check if audio is stereo
            waveform = torch.mean(waveform, dim=0, keepdim=True)  # Convert to mono
        num_frames = int(sr * audio_duration)
        waveform = waveform[:, :num_frames]  # Trim the waveform to the video duration
        audio_tensor = waveform.unsqueeze(0)

        write_video(
            outpath,
            video_array=img_tensors,
            fps=fps,
            audio_array=audio_tensor,
            audio_fps=sr,
            audio_codec="aac",
            video_codec="libx264",
        )
    else:
        write_video(
            outpath,
            video_array=img_tensors,
            fps=fps,
            video_codec="libx264",
        )

    return outpath

def create_app(name):
    app = Flask(name)
    CORS(app, resources={r"/*": {"origins": "*"}})  # This will enable CORS for all routes and all origins
    sock = Sock(app)

    print(f"cudnn version={torch.backends.cudnn.version()}")
    print(f"torch version={torch.__version__}")
    print(f"torch cuda version={torch.version.cuda}")

    if torch.cuda.is_available():
        device = "cuda"
    elif torch.backends.mps.is_available():
        device = "mps"
    else:
        device = "cpu"

    return app, sock, device


# Helper method to create a safe file name
def create_safe_file_name(file_name):
    # Replace special characters in the file name to avoid OSError
    return re.sub(r'[\\/*?:"<>|]', "", file_name)

def count_params(model, verbose=False):
    total_params = sum(p.numel() for p in model.parameters())
    if verbose:
        print(f"{model.__class__.__name__} has {total_params*1.e-6:.2f} M params.")
    return total_params


def check_istarget(name, para_list):
    """ 
    name: full name of source para
    para_list: partial name of target para 
    """
    istarget=False
    for para in para_list:
        if para in name:
            return True
    return istarget


def instantiate_from_config(config):
    if not "target" in config:
        if config == '__is_first_stage__':
            return None
        elif config == "__is_unconditional__":
            return None
        raise KeyError("Expected key `target` to instantiate.")
    return get_obj_from_str(config["target"])(**config.get("params", dict()))


def get_obj_from_str(string, reload=False):
    module, cls = string.rsplit(".", 1)
    if reload:
        module_imp = importlib.import_module(module)
        importlib.reload(module_imp)
    return getattr(importlib.import_module(module, package=None), cls)


def load_npz_from_dir(data_dir):
    data = [np.load(os.path.join(data_dir, data_name))['arr_0'] for data_name in os.listdir(data_dir)]
    data = np.concatenate(data, axis=0)
    return data


def load_npz_from_paths(data_paths):
    data = [np.load(data_path)['arr_0'] for data_path in data_paths]
    data = np.concatenate(data, axis=0)
    return data   


def resize_numpy_image(image, max_resolution=512 * 512, resize_short_edge=None):
    h, w = image.shape[:2]
    if resize_short_edge is not None:
        k = resize_short_edge / min(h, w)
    else:
        k = max_resolution / (h * w)
        k = k**0.5
    h = int(np.round(h * k / 64)) * 64
    w = int(np.round(w * k / 64)) * 64
    image = cv2.resize(image, (w, h), interpolation=cv2.INTER_LANCZOS4)
    return image


def setup_dist(args):
    if dist.is_initialized():
        return
    torch.cuda.set_device(args.local_rank)
    torch.distributed.init_process_group(
        'nccl',
        init_method='env://'
    )


def download_model_if_not_exists(model_url, model_cache_directory, model_filename, max_attempts=3, headers=None):
    """
    Downloads a model file from a given URL to a specified directory with a specified filename if it does not already exist.
    Args:
    - model_url (str): URL of the model file to download.
    - model_cache_directory (str): Directory where the model file should be saved.
    - model_filename (str): Filename to save the model file as.
    - max_attempts (int): Maximum number of download attempts. Default is 3.
    - headers (dict): Optional headers to use for the request.
    """
    if not headers:
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3'}

    full_path = os.path.join(model_cache_directory, model_filename)
    
    if os.path.exists(full_path) and os.path.getsize(full_path) > 0:
        print(f"Model file already exists at {full_path}. Skipping download.")
        return full_path

    os.makedirs(model_cache_directory, exist_ok=True)

    attempt = 0
    success = False
    while attempt < max_attempts and not success:
        try:
            response = requests.get(model_url, headers=headers, stream=True)
            response.raise_for_status()
            total_size_in_bytes = int(response.headers.get('content-length', 0))
            if total_size_in_bytes == 0:
                print("Received a 0-byte file, retrying...")
                attempt += 1
                continue

            progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
            block_size = 1024  # 1 Kibibyte
            with open(full_path, 'wb') as file:
                for chunk in response.iter_content(block_size):
                    file.write(chunk)
                    progress_bar.update(len(chunk))
            progress_bar.close()

            if os.path.getsize(full_path) > 0:
                print("\nModel downloaded successfully.")
                success = True
            else:
                print("Downloaded file is 0 bytes, retrying...")
                attempt += 1
        except Exception as e:
            print(f"An error occurred: {e}")
            attempt += 1

    if not success:
        raise Exception("Failed to download model after multiple attempts.")
    return full_path




import asyncio
import datetime
import json
import os
import signal
import time
import click

from cli.fingerprint import get_fingerprint
from .lib import create_venv, get_and_download_plugin, get_current_user, get_inference_job, get_my_workers, install_requirements, install_torch_requirements, login_with_token, post_register_worker, process_inference_job, run_plugin

@click.group()
def cli():
    pass

@click.command(help="Share my AI CLI - Help")
def help():
    click.echo(f'Help:')
    click.echo(f'  - help: Show this help message')
    return

@click.command(help="Login with a token")
@click.option('--token', required=True, help='Your login token')
def login(token):
    logged_in_user = login_with_token(token)
    if not logged_in_user:
        click.echo('Login failed, please try again.')
        return
    click.echo(f'Logged in successfully as {logged_in_user["name"]} - {logged_in_user["id"]}')
    return

@click.command(help="Get the current user")
def current_user():
    user = get_current_user()
    formatted_json = json.dumps(user, indent=4, sort_keys=True)
    click.echo(f'Current user:\n{formatted_json}')
    return

@click.command(help="Get the current machines fingerprint")
def fingerprint():
    fingerprint_bytes = asyncio.run(get_fingerprint())
    fingerprint_str = fingerprint_bytes.hex()
    click.echo(f'Fingerprint: {fingerprint_str}')
    return

@click.command(help="Register new worker")
@click.option('--plugin_id', required=True, help='The plugin id to work on')
def register_worker(plugin_id):
    worker = post_register_worker(plugin_id)
    formatted_json = json.dumps(worker, indent=4, sort_keys=True)
    click.echo(f'Worker:\n{formatted_json}')
    return

@click.command(help="Get my workers")
def my_workers():
    workers = get_my_workers()
    formatted_json = json.dumps(workers, indent=4, sort_keys=True)
    click.echo(f'Workers:\n{formatted_json}')
    return

@click.command(help="Run worker")
@click.option('--plugin_id', required=True, help='The plugin id to work on')
def run_worker(plugin_id):
    # first, register a worker
    worker = post_register_worker(plugin_id)
    formatted_json = json.dumps(worker, indent=4, sort_keys=True)
    click.echo(f'Worker:\n{formatted_json}')
    # then, download the plugin
    plugin, plugin_path = get_and_download_plugin(plugin_id)
    click.echo(f'Plugin downloaded to {plugin_path}')
    # then, create a venv for the plugin
    venv_path = create_venv(plugin_path)
    click.echo(f'Venv created at {venv_path}')
    # then, install the plugin requirements
    install_torch_requirements(venv_path, plugin['torchRequirements'])
    install_requirements(venv_path, plugin['requirements'])
    click.echo(f'Requirements installed')
    # then, run the plugin
    pid = run_plugin(venv_path, plugin_path, plugin['name'])
    try:
        while True:
            click.echo(f'Checking for jobs - {datetime.datetime.now()}...')
            job = get_inference_job(worker['id'])
            if not job:
                click.echo(f'No job found - {datetime.datetime.now()}...')
                time.sleep(5)
                continue
            click.echo(f'Job found: {job}')
            # process it 
            process_inference_job(job['params']['prompt'], job['params']['image'], job['params'], plugin['params'], job['id'])
            time.sleep(5)
    except KeyboardInterrupt:
        click.echo(f'Stopping plugin - {datetime.datetime.now()}...')
        os.kill(pid, signal.SIGTERM)
    click.echo(f'Plugin stopped - {datetime.datetime.now()}...')
    return

cli.add_command(help)
cli.add_command(login)
cli.add_command(current_user)
cli.add_command(fingerprint)
cli.add_command(register_worker)
cli.add_command(my_workers)
cli.add_command(run_worker)






import base64
import gc
import io
import json
import os
import sys
from PIL import Image
from .utils import save_different_params, save_image

# generator method to make websocket creation easier
# keeps us to standard format
# we check passed in params to know what pipes to look for and what data to expect
def create_gen_img_websocket(sock, plugin_name, pipe_dict, default_params = {
    "prompt": "",
    "negative_prompt": "",
    "image": "",
    "inference_steps": 40,
    "guidance_scale": 5.0,
    "strength": 1.0,
    "batch_size": 1,
    "height": 512,
    "width": 512
}, forward_pass=None):
    def generate_img_common(ws):
        while True:
            data_str = ws.receive()
            if not data_str:
                continue
            try:
                data = json.loads(data_str)
            except json.JSONDecodeError:
                data = {'prompt': data_str}
            params = {**default_params, **data}
           
            if not params.get('prompt') and not params.get('image'):
                ws.send(json.dumps({'error': 'Either prompt or image is required'}))
                continue
            
            pipe = pipe_dict.get('pipeline', None)
            refiner = pipe_dict.get('refiner', None)
          
            # Check if the image is sent as base64 or as a file object
            if params.get('image'):
                image = None
                # First, check if it looks like a file path or object
                if any(ext in params['image'] for ext in ['.jpg', '.jpeg', '.png', '.gif']):
                    try:
                        image = Image.open(params['image'])
                    except FileNotFoundError:
                        ws.send(json.dumps({'error': 'Image file not found'}))
                    except Exception as e:
                        ws.send(json.dumps({'error': 'Error opening image file'}))
                else:
                    # If it doesn't look like a file, attempt to decode as base64
                    try:
                        image_data = base64.b64decode(params['image'].split(',')[-1])  # Handle potential data URI scheme
                        image = Image.open(io.BytesIO(image_data))
                    except Exception as e:
                        ws.send(json.dumps({'error': 'Invalid or unrecognized image data'}))
            else:
                image = None

            def callback(pipe, step_index, timestep, callback_kwargs):
                step = step_index + 1
                ws.send(json.dumps({
                    'step': step,
                    'inference_steps': params.get('inference_steps', 40),
                    'progress': f'{(step/params.get("inference_steps", 40))*100}'
                }))
                return callback_kwargs
            
            try:
                # Generate images based on the presence of an initial image or just a prompt
                generation_args = {
                    "prompt": params.get('prompt'),
                    "negative_prompt": params.get('negative_prompt'),
                    "num_images_per_prompt": params.get('batch_size', 1),
                    "num_inference_steps": params.get('inference_steps', 40),
                    "guidance_scale": params.get('guidance_scale', 5.0),
                    "height": params.get('height', 512),
                    "width": params.get('width', 512),
                    "eta": params.get('eta', 0.3),
                    "callback_on_step_end": callback
                }
                if image is not None:
                    generation_args.update({"image": image, "strength": params.get('strength', 1.0)})
                result = forward_pass(pipe, **generation_args) if forward_pass else pipe(**generation_args)
            except Exception as e:
                ws.send(json.dumps({'error': str(e)}))
                continue

            # Refine generated images if required
            if refiner and params.get('use_refiner', False):
                refine_args = {
                    "prompt": params.get('prompt'),
                    "negative_prompt": params.get('negative_prompt'),
                    "image": result.images,
                    "num_inference_steps": params.get('inference_steps', 40),
                    "height": params.get('height', 512),
                    "width": params.get('width', 512),
                    "callback_on_step_end": callback
                }
                result = refiner(**refine_args)

            # if the params are different from our server config, save them to the server config
            save_different_params(plugin_name, params)
            # Save generated images to disk and collect their file paths
            file_paths = []
            for idx, image in enumerate(result.images, start=1):
                out_path = save_image(image, "png", f"{params.get('prompt')}_{idx}")
                full_path = os.path.abspath(out_path)
                file_paths.append(full_path)

            # Return the array of full file paths as a response
            ws.send(json.dumps({
                'file_paths': file_paths
            }))

            # Free GPU memory and RAM after each call
            del result, image, pipe, refiner
            if 'torch' in sys.modules:
                import torch
                torch.cuda.empty_cache()
                
            gc.collect()

    @sock.route('/ws/generate/image')
    def generate_img(ws):
        generate_img_common(ws)

    @sock.route('/ws/generate')
    def generate(ws):
        generate_img_common(ws)


import datetime
import json
import os
import platform
import time
import psutil
import torch
from flask import jsonify, request
from .utils import check_for_saved_config


def create_defaults(app, sock, plugin_name, pipes, pipe_dict, default_params, load_pipe):
    check_for_saved_config(plugin_name, pipe_dict, default_params, load_pipe)
    @app.route('/')
    def home():
        return jsonify({'message': f'Welcome to the world {plugin_name}!'})
    
    @sock.route('/ws/healthcheck')
    def healthcheck_ws(ws):
        while True:
            data_str = ws.receive()
            if data_str:
                try:
                    data = json.loads(data_str)
                except json.JSONDecodeError:
                    data = {'request': data_str}
                if data.get('request') == 'system_stats':  # if they send a message requesting system stats
                    if torch.cuda.is_available():
                        device = "cuda"
                        cuda_version = torch.version.cuda
                    elif torch.backends.mps.is_available():
                        device = "mps"
                        cuda_version = "Not applicable"
                    else:
                        device = "cpu"
                        cuda_version = "Not applicable"
                    device_info = platform.uname()
                    system_stats = {
                        "cpu_percent": psutil.cpu_percent(),
                        "virtual_memory": psutil.virtual_memory(),
                        "disk_usage": psutil.disk_usage('/')
                    }
                    current_time = datetime.datetime.now().isoformat()
                    ws.send(json.dumps({
                        "device_info": device_info,
                        "system_stats": system_stats,
                        "current_time": current_time,
                        "environment_name": plugin_name,
                        "torch_device": device,
                        "cuda_version": cuda_version,
                        "mem_info": torch.cuda.mem_get_info()
                    }))
            time.sleep(1)  # sleep for a second before checking again

    @app.route('/healthcheck')
    def healthcheck():
        if torch.cuda.is_available():
            device = "cuda"
            cuda_version = torch.version.cuda
        elif torch.backends.mps.is_available():
            device = "mps"
            cuda_version = "Not applicable"
        else:
            device = "cpu"
            cuda_version = "Not applicable"
        device_info = platform.uname()
        system_stats = {
            "cpu_percent": psutil.cpu_percent(),
            "virtual_memory": psutil.virtual_memory(),
            "disk_usage": psutil.disk_usage('/')
        }
        current_time = datetime.datetime.now().isoformat()
        return jsonify({
            "device_info": device_info,
            "system_stats": system_stats,
            "current_time": current_time,
            "environment_name": plugin_name,
            "torch_device": device,
            "cuda_version": cuda_version,
            "mem_info": torch.cuda.mem_get_info()
        })
    
    @app.route("/config")
    def get_config():
        # everytime we call get config, save the config to a json file in the model directory, so we can re-init the app with the same config
        model_dir = os.getenv("HF_HOME") or "./models" 
        filename = os.path.join(model_dir, f"{plugin_name}_server_config.json")
        with open(filename, "w") as f:
            json.dump({
                "default_params": default_params,
                "pipelines": list(pipes.keys()),
                "pipeline": pipe_dict["name"],
                "loras": pipe_dict["loras"],
                "adapter_weights": pipe_dict["adapter_weights"],
            }, f)

        return jsonify({
            "default_params": default_params,
            "pipelines": list(pipes.keys()),
            "pipeline": pipe_dict["name"],
            "loras": pipe_dict["loras"],
            "adapter_weights": pipe_dict["adapter_weights"],
        })
    
    @app.route("/pipeline", methods=['PUT'])
    def update_pipeline():
        # global pipe_dict
        if request.is_json:
            data = request.get_json()
        else:
            data = request.form
        pipeline = data.get("pipeline")
        if not pipeline in pipes:
            return jsonify({"message": "Pipeline not found"}), 404
        
        pipe_dict["pipeline"] = load_pipe(pipeline)
        pipe_dict["name"] = pipeline
        if data.get("refiner"):
            pipe_dict["refiner"] = load_pipe("refiner")
        else:
            pipe_dict["refiner"] = None

        # Reload any loaded_loras when we swap our pipeline
        if pipe_dict["loras"]:
            for lora in pipe_dict["loras"]:
                pipe_dict["pipeline"].load_lora_weights(lora['url'], weight_name=lora['weight_name'], adapter_name=lora['adapter_name'])
            pipe_dict["pipeline"].set_adapters([lora['adapter_name'] for lora in pipe_dict['loras']], adapter_weights=pipe_dict["adapter_weights"])
            pipe_dict["pipeline"].fuse_lora(adapter_names=[lora['adapter_name'] for lora in pipe_dict['loras']])

        return jsonify({"message": "Pipeline updated"})
    

def create_lora_helpers(app, pipes, pipe_dict, default_params, load_pipe):
    print("Creating lora helpers")
    
    @app.route("/loras", methods=['PUT'])
    def update_loras():
        if request.is_json:
            data = request.get_json()
        else:
            data = request.form
        loras = data.get("loras")
        if not loras:
            pipe_dict['loras'] = []
            pipe_dict["pipeline"].unfuse_lora()
            return jsonify({"message": "Loras reset"}), 200
        pipe_dict['loras'] = loras
        # Unload all adapters before loading new ones to avoid ValueError
        pipe_dict["pipeline"] = load_pipe(pipe_dict["name"])
        for lora in loras:
            print(lora)
            pipe_dict["pipeline"].load_lora_weights(lora['url'], weight_name=lora['weight_name'], adapter_name=lora['adapter_name'])
        adapter_weights = data.get("weights")
        pipe_dict["adapter_weights"] = adapter_weights
        pipe_dict["pipeline"].set_adapters([lora['adapter_name'] for lora in loras], adapter_weights=adapter_weights)
        pipe_dict["pipeline"].fuse_lora(adapter_names=[lora['adapter_name'] for lora in loras])
        return jsonify({"message": "Loras updated"})
    






import hashlib
import os
import platform
import sys
import psutil

async def get_fingerprinting_info():
    system_info = platform.uname()
    bios_info = platform.libc_ver()
    cpu_info = {
        "manufacturer": platform.processor(),
        "brand": platform.machine(),
        "speedMax": psutil.cpu_freq().max,
        "cores": psutil.cpu_count(logical=False),
        "physicalCores": psutil.cpu_count(logical=True),
    }
    mem_info = {
        "total": psutil.virtual_memory().total,
    }
    os_info_data = {
        "platform": system_info.system,
        "arch": system_info.machine,
    }
    devices = psutil.disk_partitions()
    hdds = [f"{device.device}{device.fstype}" for device in devices if not psutil.disk_usage(device.mountpoint).free == 0]

    return {
        "EOL": os.linesep,
        "endianness": "little" if sys.byteorder == "little" else "big",
        "manufacturer": system_info.node,
        "model": system_info.system,
        "serial": system_info.version,
        "uuid": system_info.machine,
        "vendor": bios_info[0],
        "biosVersion": bios_info[1],
        "releaseDate": "N/A",
        "boardManufacturer": "N/A",
        "boardModel": "N/A",
        "boardSerial": "N/A",
        "cpuManufacturer": cpu_info['manufacturer'],
        "brand": cpu_info['brand'],
        "speedMax": f"{cpu_info['speedMax']:.2f}",
        "cores": cpu_info['cores'],
        "physicalCores": cpu_info['physicalCores'],
        "socket": "N/A",
        "memTotal": mem_info['total'],
        "platform": os_info_data['platform'],
        "arch": os_info_data['arch'],
        "hdds": hdds,
    }

async def get_fingerprinting_parameters():
    info = await get_fingerprinting_info()
    return list(info.keys())

def calculate_fingerprint(parameters):
    fingerprint_string = ''.join(str(parameters[param]) for param in sorted(parameters) if param in parameters)
    fingerprint_hash = hashlib.sha512(fingerprint_string.encode())
    return fingerprint_hash.digest()

cached_fingerprints = {}

async def get_fingerprint(options=None):
    if options is None:
        options = {}
    default_parameters = await get_fingerprinting_parameters()
    only = options.get('only', default_parameters)
    except_params = options.get('except', [])
    parameters = [param for param in default_parameters if param in only and param not in except_params]
    cache_key = ''.join(parameters)
    if cache_key not in cached_fingerprints:
        info = await get_fingerprinting_info()
        selected_info = {param: info[param] for param in parameters}
        cached_fingerprints[cache_key] = calculate_fingerprint(selected_info)
    return cached_fingerprints[cache_key]



from setuptools import setup, find_packages                                     
import codecs                                                                   
import os                                                                       
import re      

def find_version(*file_paths):                                                  
    # Open in Latin-1 so that we avoid encoding errors.                         
    # Use codecs.open for Python 2 compatibility                                
    here = os.path.abspath(os.path.dirname(__file__))                           
    with codecs.open(os.path.join(here, *file_paths), 'r', 'latin1') as f:      
        version_file = f.read()                                                 

    # The version line must have the form                                          
    # __version__ = 'ver'                                                       
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",            
                              version_file, re.M)                               
    if version_match:                                                           
        return version_match.group(1)                                           
    raise RuntimeError("Unable to find version string.")                                                                 

long_description = """
Reusable utilities for ShareMyAI. For more information, visit [ShareMyAI](https://sharemy.ai).
"""

setup(                                                                          
    name="sharemyai_utils",                                                             
    version=find_version('sharemyai_utils', '__init__.py'),                             
    description="Reusable utils for sharemyai plugins",                    
    long_description=long_description,                                          
    url='http://github.com/catswithkeyboards',                                   
    author='Cats with Keyboards',                                                     
    author_email='tech@catswithkeyboards.com',                                     
    license='',                                    
    classifiers=[                                                               
        'Development Status :: 3 - Alpha',       
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'License :: Free for non-commercial use'
    ],                                                                          
    keywords='sharemyai utils',                                                  
    packages=find_packages(exclude=["tests*", "*.env", "scripts"]),                                 
    install_requires=[   
        "click",
        "psutil==5.9.8",
        "pillow",
        "flask==3.0.2",
        "flask_cors==4.0.0",
        "flask-sock==0.7.0",
        "httptools==0.6.1",
        "opencv-python==4.9.0.80",
        "websockets"                                                 
    ],   
    entry_points={
        "console_scripts": [
            "sharemyai = cli:cli"
        ]
    },                                                                       
)      